@@ -13,7 +13,7 @@ use Data::Dumper;
 
 use IO::Uncompress::Gunzip qw(gunzip $GunzipError) ;
 
-our $VERSION = "0.044";
+our $VERSION = "0.045";
 
 my %INFO_ROUTES = (
     version => '/version',
@@ -41,7 +41,9 @@ sub new {
             sprintf 'http://%s:%d', @params{qw<host port>};
     }
 
-    my $http_tiny = HTTP::Tiny->new();
+    my $http_tiny = HTTP::Tiny->new(
+            defined $params{'timeout'} ? ( timeout => $params{'timeout'} ) : (),
+        );
 
     my $strict_mode = 0;
     if ($params{strict_mode}) {
@@ -49,7 +51,7 @@ sub new {
             $strict_mode = $params{strict_mode};
         }
         else {
-            die "Invalid value. Strict mode can have only one of the following values: [undef, 0, 1]";    
+            die "Invalid value. Strict mode can have only one of the following values: [undef, 0, 1]";
         }
     }
 
@@ -73,29 +75,15 @@ sub new {
 sub list {
     my $self = shift;
 
-    $self->{last_error} = undef;
-
     my $uri = $self->{service} . $INFO_ROUTES{list};
 
-    my $http = $self->{http_tiny} ||=  HTTP::Tiny->new();
-
-    my $rs = $http->get($uri, {
+    my $rs = $self->{http_tiny}->get($uri, {
          headers => {
              'Accept' => 'application/json',
          }
     });
 
-    if ( ! $rs->{success} ) {
-
-        $self->{last_error} = _extract_error_tiny($uri, $rs);
-
-        if ($self->{strict_mode}) {
-            die "request error: " . Dumper($self->{last_error});
-        }
-        else {
-            return undef;
-        }
-    }
+    return if $self->_handle_error( $uri, $rs );
 
     my $response = decode_json($rs->{content});
 
@@ -115,29 +103,15 @@ sub list {
 sub version {
     my $self = shift;
 
-    $self->{last_error} = undef;
-
     my $uri = $self->{service} . $INFO_ROUTES{version};
 
-    my $http = $self->{http_tiny} ||= HTTP::Tiny->new();
-
-    my $rs = $http->get($uri, {
+    my $rs = $self->{http_tiny}->get($uri, {
          headers => {
              'Accept' => 'application/json',
          }
     });
 
-    if ( ! $rs->{success} ) {
-
-        $self->{last_error} = _extract_error_tiny($uri, $rs);
-
-        if ($self->{strict_mode}) {
-            die "request error: " . Dumper($self->{last_error});
-        }
-        else {
-            return undef;
-        }
-    }
+    return if $self->_handle_error( $uri, $rs );
 
     my $response = decode_json($rs->{content});
 
@@ -170,15 +144,22 @@ sub get {
     my $self   = shift;
     my $params = shift;
 
-    $self->{last_error} = undef;
-
     my $get_urls = _build_get_uri($params);
 
     my @result = ();
     foreach my $url (@$get_urls) {
-        my $rows = $self->_get_tiny($url->{url});
-        push @result, @$rows
-            if ($rows and @$rows);
+
+        if ( my $rows = $self->_get_tiny( $url->{url} ) ){
+
+            push @result, @$rows;
+
+        } else {
+
+            # we allow for some keys to be missing but fail on other errors
+            return unless $self->{last_error} && ( $self->{last_error}->{type} // '' ) eq '404';
+
+        }
+
     }
 
     return \@result;
@@ -192,27 +173,15 @@ sub _get_tiny {
 
     my $url = $self->{service} . $uri;
 
-    my $http = $self->{http_tiny} ||= HTTP::Tiny->new();
-    
-    my $rs = $http->get($url, {
+    my $rs = $self->{http_tiny}->get($url, {
         headers => {
             'Accept' => 'application/json',
             'Accept-Encoding'   => 'gzip',
         }
     });
 
-    if ( ! $rs->{success} ) {
+    return if $self->_handle_error( $uri, $rs, [ '404' ] );
 
-        $self->{last_error} = _extract_error_tiny($url, $rs);
-
-        if ($self->{strict_mode}) {
-            die "request error: " . Dumper($self->{last_error});
-        }
-        else {
-            return undef;
-        }
-    }
-    
     _maybe_decompress( $rs );
     my $response = decode_json( $rs->{content} );
 
@@ -252,8 +221,6 @@ sub multiget {
     my $self  = shift;
     my $query = shift;
 
-    $self->{last_error} = undef;
-
     my $where = $query->{where};
     unless ($where->{key_in} && @{$where->{key_in}}) {
         $self->{last_error} = {
@@ -267,21 +234,20 @@ sub multiget {
     my $multiget_urls = _build_multiget_uri($query);
 
     my @result = ();
+
     foreach my $url (@$multiget_urls) {
-        my $rows = $self->_multiget_tiny($url->{url});
-        if ($self->{last_error}) {
-            if ($self->{last_error}->{type} eq '404') {
 
-                # if some of the keys don't exists, just ignore it
-                $self->{last_error} = undef;
+        if ( my $rows = $self->_multiget_tiny( $url->{url} ) ){
+
+            push @result, @$rows;
+
+        } else {
+
+            # we allow for some keys to be missing but fail on other errors
+            return unless $self->{last_error} && ( $self->{last_error}->{type} // '' ) eq '404';
 
-            } else {
-                #some other error in the middle, fail the whole thing
-                return [];
-            }
         }
-        push @result, @$rows
-            if ($rows and @$rows);
+
     }
 
     return \@result;
@@ -298,28 +264,17 @@ sub _multiget_tiny {
 
     my $url = $self->{service} . $uri;
 
-    my $http = $self->{http_tiny} ||= HTTP::Tiny->new();
-
     my $data_format = 'application/json';
 
-    my $rs = $http->get($url, {
+    my $rs = $self->{http_tiny}->get($url, {
         headers => {
             'Accept' => $data_format,
             'Accept-Encoding'   => 'gzip',
         }
     });
 
-    if ( ! $rs->{success} ) {
-
-        $self->{last_error} = _extract_error_tiny($url, $rs);
-
-        if ($self->{strict_mode}) {
-            die "request error: " . Dumper($self->{last_error});
-        }
-        else {
-            return undef;
-        }
-    }
+    # allow items to be missing for multiget
+    return if $self->_handle_error( $uri, $rs, [ '404' ] );
 
     _maybe_decompress( $rs );
     my $response = decode_json($rs->{content});
@@ -356,7 +311,7 @@ sub _multiget_tiny {
 #              {
 #                   column    => "$family:$name",
 #                   value     => "$value",
-#                   timestamp => $timestamp # <- optional (override HBase timestamp) 
+#                   timestamp => $timestamp # <- optional (override HBase timestamp)
 #               },
 #              ...,
 #              { column => "$family:$name", value => "$value" },
@@ -371,8 +326,6 @@ sub put {
     my $self    = shift;
     my $command = shift;
 
-    $self->{last_error} = undef;
-
     # at least one valid record
     unless ($command->{table} &&
             (defined $command->{changes}->[0]->{row_key}) &&
@@ -461,9 +414,7 @@ sub put {
     my $route = '/' . uri_escape($table) . '/false-row-key';
     my $uri = $self->{service} . $route;
 
-    my $http = $self->{http_tiny} ||= HTTP::Tiny->new();
-
-    my $rs = $http->request('PUT', $uri, {
+    my $rs = $self->{http_tiny}->request('PUT', $uri, {
         content => $JSON_Command,
         headers => {
             'Accept'       => 'application/json',
@@ -471,17 +422,7 @@ sub put {
         },
     });
 
-    if ( ! $rs->{success} ) {
-
-        $self->{last_error} = _extract_error_tiny($uri, $rs);
-        
-        if ($self->{strict_mode}) {
-            die "request error: " . Dumper($self->{last_error});
-        }
-        else {
-            return undef;
-        }
-    }
+    return !$self->_handle_error( $uri, $rs );
 
 }
 
@@ -501,7 +442,7 @@ sub delete {
     my $key    = delete $attr->{key};
     my $family = delete $attr->{family};
     my $column = delete $attr->{column};
-    my ($route, $http, $rs, $url);
+    my ($route, $rs, $url);
 
     die "Table name required" if(!$table);
     die "Row key required" if(!$key);
@@ -515,29 +456,15 @@ sub delete {
         $route .= sprintf(":%s", $column) if($column);
     }
 
-    $http = $self->{http_tiny} ||= HTTP::Tiny->new();
     $url = sprintf("%s%s", $self->{service}, $route);
-    $rs = $http->delete($url, {
+    $rs = $self->{http_tiny}->delete($url, {
         headers => {
             'Accept' => 'application/json',
         }
     });
 
-    if($rs->{success}) {
-       $self->{last_error} = undef;
-       return 1;
-    }
-    else {
+    return !$self->_handle_error( $url, $rs );
 
-        $self->{last_error} = _extract_error_tiny($url, $rs);
-        
-        if ($self->{strict_mode}) {
-            die "request error: " . Dumper($self->{last_error});
-        }
-        else {
-            return 0;
-        }
-    }
 }
 
 # -------------------------------------------------------------------------
@@ -643,7 +570,7 @@ sub _build_multiget_uri {
         foreach my $mget_url (@multiget_urls) {
             my $versions = $query->{versions};
             my $versions_url_part = "v=$versions";
-            
+
             $mget_url->{url} .= '&' . $versions_url_part;
         }
     }
@@ -651,6 +578,31 @@ sub _build_multiget_uri {
     return \@multiget_urls;
 }
 
+# -------------------------------------------------------------------------
+# Handles the error response:
+# 1) Replaces $self->{last_error} with a one parsed from the response (that can be undef)
+# 2) If there is a error, returns true in non-strict mode. In strict mode dies on error unless
+#    its type is given not-to-die otherwise returns true
+#
+sub _handle_error {
+
+    my ( $self, $uri, $response, $not_to_die ) = @_;
+
+    if ( my $error = $self->{last_error} = _extract_error_tiny($uri, $response) ) {
+
+        if ( $self->{strict_mode} ) {
+
+            die "request error: " . Dumper( $error ) unless $error->{type}
+                                                        and $not_to_die
+                                                        and grep { $_ eq $error->{type} } @$not_to_die ;
+
+        }
+
+        return 1;
+    }
+
+}
+
 # -------------------------------------------------------------------------
 # parse error
 #
@@ -663,35 +615,48 @@ sub _extract_error_tiny {
 
     my $detailed_error_info = {reason => $res->{reason}, content => $res->{content}, status => $res->{status}};
 
-    if (!$res->{status}) {
-         return {
-            type => 'Unknown',
-            info => 'No status in response',
-            uri  => $uri,
-            http_response => $res,
-        };
-    }
+    if ( my $http_status = $res->{status} ){
 
-    if ($res->{status} and $res->{status} == 404) {
-        return {
-            type => '404',
-            info => 'A subset of keys you\'ve requested was not found. Or: no data has been written, if you were writing',
-            guess => 'Non-existing table, subset of keys or an exceeded quota?', #at the time of this writing, HBase REST servers send you a 404 when you're over quota, reading. This requires a fix on HBase side, no way to work around this here.
-            uri => $uri
-        };
-    }
+        if ( $http_status == 404 ) {
+            return {
+                type => '404',
+                info => 'A subset of keys you\'ve requested was not found. Or: no data has been written, if you were writing',
+                guess => 'Non-existing table, subset of keys or an exceeded quota?', #at the time of this writing, HBase REST servers send you a 404 when you're over quota, reading. This requires a fix on HBase side, no way to work around this here.
+                uri => $uri
+            };
+        } elsif ( $http_status == 599 ) {
 
-    if ($res->{status} and $res->{status} == 503){
-        my @lines = split /\n/, $res->{content};
+            return {
+                    type => '599',
+                    info => 'Timeout',
+                    uri => $uri,
+                    error_details => $detailed_error_info
+                };
 
-        foreach my $line (@lines)
-        {
-            if (index($line, 'ThrottlingException') != -1)
+        } elsif ( $http_status == 503 ){
+
+            my @lines = split /\n/, $res->{content};
+
+            foreach my $line (@lines)
             {
-               # TODO: type => 404 above, how do we make it consistent?
-               return {type => '503', exception => 'QuotaExceededException', info => $line, uri => $uri, error_details => $detailed_error_info};
+                if (index($line, 'ThrottlingException') != -1)
+                {
+                   # TODO: type => 404 above, how do we make it consistent?
+                   return {type => '503', exception => 'QuotaExceededException', info => $line, uri => $uri, error_details => $detailed_error_info};
+                }
             }
+
         }
+
+    } else {
+
+        return {
+                type    => 'Unknown',
+                info    => 'No status in response',
+                uri     => $uri,
+                http_response => $res,
+            };
+
     }
 
     my $msg;
@@ -719,7 +684,7 @@ sub _extract_error_tiny {
 
 sub _maybe_decompress {
     my $rs = shift;
-   
+
     if (    exists $rs->{headers}
             && exists $rs->{ headers }->{ 'content-encoding' }
             && $rs->{ headers }->{ 'content-encoding' } eq 'gzip' ) {
@@ -728,7 +693,7 @@ sub _maybe_decompress {
         gunzip \$content => \$content_decompressed,
             MultiStream => 1, Append => 1, TrailingData => \$scalar
         or die "gunzip failed: $GunzipError\n";
-    
+
         $rs->{content} = $content_decompressed;
     }
 }
@@ -788,7 +753,7 @@ A simple rest client for HBase.
 Constructor. Cretes an hbase client object that is used to talk to HBase.
 
     my $hostname = "hbase-restserver.mydomain.com";
-    
+
     my $hbase = HBase::JSONRest->new(host => $hostname);
 
 =head2 get
@@ -971,7 +936,7 @@ Returns a list of tables available in HBase
               {
                 'name' => 'very_big_table'
               },
-              ..., 
+              ...,
               {
                 'name' => 'super_big_table'
               }
@@ -988,10 +953,10 @@ Information on error is stored in hbase object under key last error:
         },
     });
     if ($hbase->{last_error}) {
-        # handle error    
+        # handle error
     }
     else {
-        # process records    
+        # process records
     }
 
 =head1 VERSION
